/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

#include <errno.h>

#include "eclib.h"

short ZC_lineno, ZC_pos, ZC_previous_pos;
char *ZC_filename;
FILE *ZC_file;
char *ZC_progname;
int (*ZC_fn_uncache)();

extern int (*ZC_fn_eprint_exit)();

eprint(s, stuff)
char *s;
struct { long a[8]; } stuff;
{
char bailflag;

if(ZC_filename && ZC_lineno)
fprintf(stderr, "%s[%d]: ",
strcmp(ZC_filename, "-") ? ZC_filename : "stdin",
ZC_pos ? ZC_lineno : ZC_lineno-1);
else if(ZC_progname) fprintf(stderr, "%s: ", ZC_progname);
bailflag = -1;
if(*s == '@') { ++s; bailflag = 1; fprintf(stderr, "disaster, "); }
if(*s >= '0' && *s <= '9') { bailflag = *s - '0'; ++s; }
fprintf(stderr, s, stuff);
fprintf(stderr, "\n");
if(bailflag >= 0) exit(bailflag);
if(ZC_fn_eprint_exit)
(*ZC_fn_eprint_exit)(1);
} /* eprint */

/* standard unix calls, protected */

FILE *efopen(name, mode)
char *name, *mode;
{
FILE *f;

if(name[0] == '-' && name[1] == 0) {
if(*mode == 'r') return stdin;
if(*mode == 'w' || *mode == 'a') return stdout;
}

f = fopen(name, mode);
if(!f && errno == EMFILE && ZC_fn_uncache) {
(*ZC_fn_uncache)();
f = fopen(name, mode);
}
if(f) return f;
if(*mode == 'r') eprint("cannot open %s", name);
else if(*mode == 'w' || *mode == 'a') eprint("cannot create or write to %s", name);
else eprint("@calling fopen() with invalid mode %s", mode);
return 0;
} /* efopen */

char *emalloc(n)
{
register char *s;
char *malloc();

if(! (s =  malloc(n)))
eprint("1error allocating %u bytes", n);
return s;
} /* emalloc */

eopen(name, mode, perms)
char *name;
{
int fd;
fd = open(name, mode, perms);
if(fd < 0 && errno == EMFILE && ZC_fn_uncache) {
(*ZC_fn_uncache)();
fd = open(name, mode, perms);
}
if(fd < 0)
eprint("cannot open %s", name);
return fd;
} /* eopen */

eread(fd, buf, n)
char *buf;
unsigned n;
{
unsigned rc;
rc = read(fd, buf, n);
if(rc && rc != n)
eprint("cannot read %d[%u]", fd, n);
return rc;
} /* eread */

ewrite(fd, buf, n)
char *buf;
unsigned n;
{
unsigned rc;
rc = write(fd, buf, n);
if(rc != n)
eprint("cannot write %d[%u]", fd, n);
return rc;
} /* ewrite */

long elseek(fd, offset, mode)
long offset;
{
long l, lseek();
l = lseek(fd, offset, mode);
if(l < 0)
eprint("cannot lseek %d[%ld]", fd, offset);
return l;
} /* elseek */

long filesize(fd) { return elseek(fd, 0L, 2); }

/* extend the length of a file, specify fill character */
file_extend(fd, offset, value)
short fd;
long offset;
long value;
{
long base = filesize(fd);
char *s;
int n;

if(offset > base) {
s = emalloc(4096);
for(n=0; n<1024; ++n)
((long *)s)[n] = value;

while(offset > base) {
if(offset-base > 4096) n = 4096;
else n = offset - base;
ewrite(fd, s, n);
base += n;
} /* file extended */
free(s);
}
} /* file_extend */

erename(file1, file2)
char *file1, *file2;
{
unlink(file2);
#ifdef MSDOS
if(rename(file1, file2))
#else
if(link(file1, file2) ||
unlink(file1))
#endif
eprint("cannot rename %s to %s", file1, file2);
} /* erename */

erm(file)
char *file;
{
if(unlink(file))
eprint("cannot remove file %s", file);
} /* erm */

/* When writing a critical file */
eputc(c, f)
FILE *f;
{
if(putc(c,f) < 0)
eprint("Error writing to output file, check disk space");
} /* eputc */

int (*ZC_fn_eclib_error_exit)();

eclib_error(e_type, n)
{
char *s;
static char bye[] = ".\n\tabandoning program at this point.";
static char e_act[] = {-1,
1,1,1,0,1,2,2,2,2,3,1,1,1,-1,-1,-1,-1,-1,-1,
1,4,1,6,1,5,5,1,5,5,
5,1,1,1,-1,5,5,5,5,5,1,5};
static char *emsgs[] = {0,
"newline encountered before the string is closed",
"string or multi-precision number is too long, limit %d characters",
"@cannot fseek back to known good point",
0,
"too many files opened concurrently, limit %d",
"1EOF is not expected in this context, a gramatical structure is incomplete%s",
"1EOF encountered within a comment%s",
"1EOF encountered within a string%s",
"1nested comments, earlier comment is probably not closed%s",
0,
"constant exceeds the precision of a long",
"character 0x%x unexpected",
"file or directory name too long, limit %d characters",
0,0,0,0,0,0,
"@calling set_alev(%d), level out of range",
"@trying to allocate %d bytes, limit %d",
"too much total memory allocated",
"node %s consumes too much memory, limit %ldK",
"@malloc failure errno %d, check memory space",
"@node %s cannot be allocated at level %d",
"@corrupt chain of available %s nodes",
"@null node pointer unexpected, location %d",
"@check_mem_empty finds %s nodes at level %d",
"duplicate %s in list",
"@missing link pointer in list %s",
"directories for serial lists not established",
"directories for serial lists established more than once",
"serial list structure not initialized",
"not used",
"cannot create lock file for %s",
"table %s locked by another user",
"serial number for %s is out of range",
"@the type of struct %s is incompatible with this operation",
"serial list of %s nodes exceeds 64K",
"structure is too large, limit %d bytes",
"%s is read only"
};
char ZC_alev;

s = emsgs[e_type];

switch(e_act[e_type]) {
case 1:
eprint(s, n);
break;
case 2:
eprint(s, bye);
break;
case 3:
if((n >> 8) == 10)
eprint("character %c is not a digit", n&0xff);
else
eprint("character %c is not allowed in a base %d number", n&0xff, n>>8);
break;
case 4:
eprint(s, n, BYTES_MALLOC);
break;
case 5:
eprint(s, struct_desc[n].name, ZC_alev);
break;
case 6:
eprint(s, struct_desc[n].name, struct_desc[n].memlimit);
break;
default: eprint("@default[1] %d in eprint.c", e_type);
} /* end switch */

if(ZC_fn_eclib_error_exit)
(*ZC_fn_eclib_error_exit)(e_type);
} /* eclib_error */

