;  vmath.asm:   variable length infinite precision integer arithmetic

;  struct VINT {  /* variable length integer */
;	short size;
;	short filler;
;	unsigned short value[2];  /* variable length array */
;	};
;  number is sign magnitude, where the sign is determined by the high order
;  bit of vint.size.
;  Thus, these numbers are limited in absolute value to 32767*16 bits.
;  
;  We always assume there is enough space for the result.
;  The format for division is noticeably non-standard.
;  vdiv(op1, op2, quot, mod), where quot must be large enough
;  to hold op1/op2 as expected, but mod must be as large as op1 + 2 words.
;  The algorithm requires this much extra temp area to work in.
;	Of course, mod must also be able to hold the remainder.
;
;  Warning: a%b is defined inconsistently with some compilers,
;	for negative arguments.
;
;  no magnitude overflow checks are made
;	In small memory model, malloc would fail first anyways.
;
;	Routines compatible with microsoft C version 4.0 small model.
;	Don't forget the -mx when you assemble this.
;
;  Operands may be the same structure,
;  but the result must be a different structure.


	title	vmath

VALUE_SIZE	equ 2
VALUE_OFS	equ 4

SIGN_OF	macro reg
	and reg,8000h
	endm

SIGN_CLR	macro reg
	and reg,7fffh
	endm

SIGN_SWAP	macro reg
	xor reg,8000h
	endm

FNSTART	macro fname
	public &FNAME
&fname:
	push bp
	mov bp,sp
	push ds ; run with es = ds
	pop es
	push si
	push di
	push bx
	mov si,[bp+4]
	mov di,[bp+6]
	mov bx,[bp+8]
	endm

FNEND	macro
	pop bx
	pop di
	pop si
	pop bp
	ret
	endm



;	microsoft 4.0 segments
_TEXT	segment  byte public 'CODE'
_TEXT	ends
_DATA	segment  word public 'DATA'
_DATA	ends
CONST	segment  word public 'CONST'
CONST	ends
_BSS	segment  word public 'BSS'
_BSS	ends
DGROUP	group	CONST,	_BSS,	_DATA
	assume  cs: _TEXT, ds: DGROUP, ss: DGROUP, es: DGROUP

_BSS	segment

sv_sign	dw ?
mul_ofs	dw ?
counter	dw ?
quotword label word
quotbyte db ?,?
divofs dw ?
dvsfwd	dw ? ; first word in divisor
dvslen	dw ? ; length of divisor

_BSS	ends


_TEXT	segment

vmath	proc near
;	magnitude add, called by vadd and vsub
uadd:	mov ax,[si]
	mov cx,[di]
	SIGN_CLR ax
	SIGN_CLR cx
;	make cx the smaller
	mov dx,ax
	sub dx,cx
	jae gl1
	xchg ax,cx
gl1:	;  make si and di relative to bx
	sub si,bx
	sub di,bx
	mov ax,[bx+si]
	SIGN_OF ax
	mov sv_sign,ax
	add bx,VALUE_OFS ; clears carry
	xor ax,[bx+di-VALUE_OFS]
	js usub; opposite signs, unsigned subtract
	jcxz ua2 ; zero operand
ua1:	mov ax,[bx+si]
	adc ax,[bx+di]
	mov [bx],ax
	inc bx ; bump by VALUE_SIZE
	inc bx
	loop ua1
ua2:	pushf
	or dx,dx
	jz ua3 ; operands the same length
	jns ua4
	mov si,di
	neg dx ; distroys carry
ua4:	mov cx,dx
	popf
ua5:	mov ax,[bx+si]
	adc  ax,0
	mov [bx],ax
	inc bx ; bump by VALUE_SIZE
	inc bx
	loop ua5
	pushf
ua3:	popf
	jnc ua6
;  addition overflow
	mov word ptr [bx],1
	add bx,VALUE_SIZE
ua6:	mov si,[bp+8]
	sub bx,VALUE_OFS
	sub bx,si
	shr bx,1 ; log2(VALUE_SIZE)
	or bx,sv_sign
	mov [si],bx
	ret

;	subtract magnitudes
usub:	or dx,dx
	jns us1
;  op2 is blatently larger than op1, switch them
	xchg ax,si
	xchg ax,di
	xchg ax,si
	neg dx ; distroys carry
	clc
	SIGN_SWAP sv_sign
us1:	jcxz us3 ; zero operand
us2:	mov ax,[bx+si]
	sbb ax,[bx+di]
	mov [bx],ax
	inc bx ; bump by VALUE_SIZE
	inc bx
	loop us2
us3:	pushf
	mov cx,dx
	jcxz us5 ; operands the same length
	popf
us4:	mov ax,[bx+si]
	sbb  ax,0
	mov [bx],ax
	inc bx ; bump by VALUE_SIZE
	inc bx
	loop us4
	pushf
us5:	mov si,[bp+8]
	mov cx,bx
	sub cx,VALUE_OFS
	sub cx,si
	shr cx,1 ; log2(VALUE_SIZE)
	mov [si],cx
	mov ax,sv_sign
	or[si],ax
	popf
	jnc us6
;  overflow, complement result
	SIGN_SWAP [si]
	add si,VALUE_OFS
	clc
us7:	mov ax,0
	sbb ax,[si]
	mov [si],ax
	inc si ; bump by VALUE_SIZE
	inc si
	loop us7
us6:	mov bx,[bp+8]

magset:	;	set magnitude of bx[], drop leading zeros
	mov si,bx
	add bx,VALUE_OFS
	mov cx,[si]
	SIGN_CLR cx
	jcxz mg3
	shl cx,1 ; log2(VALUE_SIZE)
	add bx,cx
	shr cx,1 ; log2(VALUE_SIZE)
mg2:	cmp word ptr [bx-2],0
	jnz mg3
	sub bx,VALUE_SIZE
	loop mg2
mg3:	SIGN_OF [si]
	sub bx,si
	sub bx,VALUE_OFS
	shr bx,1 ; log2(VALUE_SIZE)
	or [si],bx
	or bx,bx
	jnz mg1
	SIGN_CLR [si]
mg1:	mov bx,si ; preserve bx
	ret

;  variable length addition
	FNSTART _vadd
	call uadd
	jmp short vsubret

;  variable length subtraction
	FNSTART _vsub
	mov word ptr [bx],0
	cmp si,di
	jz vsubret
;  switch sign and call addition routine
	SIGN_SWAP [di]
	call uadd
	mov di,[bp+6]
	SIGN_SWAP [di]
vsubret:	FNEND

;  variable length multiplication
	FNSTART _vmul
	mov ax,[si]
	mov cx,[di]
	mov dx,cx
	xor dx,ax
	SIGN_OF dx
	SIGN_CLR ax
	SIGN_CLR cx
;  make the second argumennt smaller, reduces loop overhead
	cmp ax,cx
	jae vm1
	mov [bp+4],di
	mov [bp+6],si
	xchg ax,cx
vm1:	mov bx,cx
	add cx,ax
	or dx,cx
;  zero out result
	add cx,VALUE_OFS/2
	mov di,[bp+8]
	mov ax,0
	repz stosw
	or bx,bx
	jz vsubret  ;  multiply by 0, done
	mov bx,[bp+8]
	mov [bx],dx
	sub bx,[bp+6]
	sub bx,[bp+4]
	sub bx,VALUE_OFS
	mov mul_ofs,bx
	mov bx,[bp+6]
	mov ax,[bx]
	SIGN_CLR ax
	mov counter,ax
	add bx,VALUE_OFS
vm6:	mov si,[bp+4]
	mov cx,[si]
	SIGN_CLR cx
	add si,VALUE_OFS
	mov di,0
vm5:	mov ax,[bx]
	mul word ptr [si]
	add ax,di
	adc dx,0
	add si,mul_ofs
	add [bx+si],ax
	adc dx,0
	sub si,mul_ofs
	mov di,dx
	add si,VALUE_SIZE
	loop vm5
	add si,mul_ofs
	mov [bx+si],di
	add bx,VALUE_SIZE
	dec counter
	jnz vm6
	or di,di
	jnz vmulret
	mov bx,[bp+8]
	dec word ptr [bx]
vmulret:	FNEND

;  variable length division
	FNSTART _vdiv
	xor ax,ax
	mov [bx],ax
	mov bx,[bp+10]
	mov [bx],ax
	mov quotword,ax
	mov cx,[si]
	mov dx,[di]
	mov ax,cx
	xor ax,dx
	SIGN_OF ax
	mov sv_sign,ax
	SIGN_CLR cx
	jz vmulret
; divide by 0 simply produces 0, no error indication
	SIGN_CLR dx
	jz vmulret

;  copy dividend into mod
	mov di,[bp+10]
	add cx,VALUE_OFS/2
	repz movsw
;	the long division routine needs two extra 0 words leading the dividend
	xor ax,ax
	stosw
	stosw
	mov di,[bp+10]
	SIGN_CLR [di]

;  We make division by small numbers relatively efficient
	cmp dx,1
	jz vd3
	jmp longdiv
vd3:	;  short division here
	mov si,[bp+6]
	mov si,[si+VALUE_OFS]
	mov bx,[bp+10]
	mov cx,[bx]
	mov di,[bp+8]
	mov [di],cx
	mov word ptr [bx],1
	sub di,bx
	add bx,VALUE_OFS
	shl cx,1 ; log2(VALUE_SIZE)
	add bx,cx
	shr cx,1 ; log2(VALUE_SIZE)
	xor dx,dx
vd4:	mov ax,[bx-2]
	div si
	mov[bx+di-2],ax
	sub bx,VALUE_SIZE
	loop vd4
	mov [bx],dx ; the remainder

vdmag:
;  adjust quotient, set sign and fix magnitude
	mov bx,[bp+8]
	mov ax,sv_sign
	or [bx],ax
	pushf
	call magset

;	fix magnitude of remainder, it might be 0
	mov bx,[bp+10]
	call magset

;  if negative quotient and nonzero remainder, complement mod
	popf
	jns vdivret
	mov cx,[bx]
	jcxz vdivret ; 0 remainder

	mov si,[bp+6]
	mov dx,[si]
	SIGN_CLR dx
	mov [bx],dx ; remainder might become large again
	sub dx,cx
	sub si,bx
	add bx,VALUE_OFS
	clc
vd9:	mov ax,[bx+si]
	sbb ax,[bx]
	mov [bx],ax
	inc bx ; bump by VALUE_SIZE
	inc bx
	loop vd9
	mov cx,dx
	jcxz vd8
vd10:	mov ax,[bx+si]
	sbb ax,0
	mov [bx],ax
	inc bx ; bump by VALUE_SIZE
	inc bx
	loop vd10
vd8:	mov bx,[bp+10]
	call magset
vdivret:	FNEND

longdiv:
;  long division starts here
;	unlike the other routines, long division is byte oriented.
;	when we use the divide instruction, only the high byte can be trusted,
;	and even this may be off by one.
;	determine the lengths of the numbers in bytes.
	mov dvslen,dx
	mov di,dx
	shl di,1 ; log2(VALUE_SIZE)
	mov bx,[bp+6]
	cmp byte ptr [bx+di+VALUE_OFS-1],0
	jnz vl1
	dec di
vl1:	mov ax,[bx+di+VALUE_OFS-2]
	mov dvsfwd,ax
	mov bx,[bp+10]
	mov si,[bx]
	shl si,1 ; log2(VALUE_SIZE)
	cmp byte ptr [bx+si+VALUE_OFS-1],0
	jnz vl2
	dec si
vl2:	mov divofs,si
	sub si,di
	jae vl8
	jmp vdmag ; dividend smaller than the divisor
vl8:	inc si
	mov counter,si
	mov bx,[bp+8]
	inc si
	shr si,1 ; byte/word ratio
	mov [bx],si
	jc vl3
	shl si,1 ; byte/word ratio
	mov byte ptr [bx+si+VALUE_OFS-1],0
vl3:

;	loop, compute each byte of the quotient
vl4:	;	set dx.ax to top words of dividend
	mov bx,[bp+10]
	add bx,VALUE_OFS
	mov di,divofs
	dec divofs
	mov ax,[bx+di-2]
	mov dx,[bx+di]
	div dvsfwd
;	it is remotely possible to generate a quotient larger than one byte
;	analogous to 1000/101
	or ah,ah
	jz vl5
	dec ax ; too large, try 255, not 256
vl5:	mov quotbyte,al
	mov cx,dvslen
	add bx,counter
	dec bx
	mov si,[bp+6]
	add si,VALUE_OFS
	sub si,bx
	xor di,di

;	we assume the processor can access words on odd byte boundaries
vl7:	mov ax,[bx+si]
	mul quotword
	add ax,di
	mov di,dx
	adc di,0
	sub [bx],ax
	adc di,0
	add bx,VALUE_SIZE
	loop vl7
	sub [bx],di
	jnc quotok
;	underflow, quotbyte was too high
	dec quotbyte
	mov bx,[bp+10]
	add bx,VALUE_OFS-1
	add bx,counter
	mov cx,dvslen
	clc
vl6:	mov ax,[bx+si]
	adc [bx],ax
	inc bx ; bump by VALUE_SIZE
	inc bx
	loop vl6
	adc word ptr [bx],0
quotok:

	mov bx,[bp+8] ; set byte in the quotient
	mov si,counter
	mov al,quotbyte
	mov [bx+si+VALUE_OFS-1],al
	dec counter
	jnz vl4

	jmp vdmag

vmath	endp

_TEXT	ends

	end
