/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* merge.c: merge people together */

#include "structs.h"

#define DIFFBUFSIZE 100

extern int mfd;
extern long foldincount;

static ia_flag; /* are we running interactive? */
static char name1[NAMELEN], name2[NAMELEN];
static long mergecount, mergepromptcount, mergeyescount;
static long *diffbuf;
static short dblen, dbnext;

static diffbufset()
{
if(!diffbuf)
diffbuf = (long *)emalloc(8*DIFFBUFSIZE);
dblen = dbnext = 0;
mergecount = mergepromptcount = mergeyescount = 0;
} /* diffbufset */

static diffbufclear()
{
if(diffbuf) free(diffbuf);
diffbuf = 0;
mergestats();
} /* diffbufclear */

static diffbufput(prs1, prs2)
long prs1, prs2;
{
diffbuf[dbnext] = prs1;
diffbuf[dbnext+1] = prs2;
dbnext += 2;
if(dbnext > dblen) dblen = dbnext;
if(dbnext == 2*DIFFBUFSIZE) dbnext = 0;
} /* diffbufput */

static indiffbuf(prs1, prs2)
long prs1, prs2;
{
int i;
for(i=0; i<dblen; i += 2)
if(diffbuf[i] == prs1 && diffbuf[i+1] == prs2) return 1;
return 0;
} /* indiffbuf */

static mergestats()
{
long newprs;
struct person *p;
char *s;
FILE *umf;

printf("%ld pairs merged automatically.\n", mergecount - mergeyescount);
printf("%ld pairs merge with your assistance.\n", mergeyescount);
printf("%ld pairs declared different (with your help).\n", mergepromptcount - mergeyescount);
if(mfd >= 0 && foldincount > mergecount) {
printf("%ld people folded in, but not merged.\n", foldincount - mergecount);
if(get_yn("Shall I list these new unmatched individuals in a file for your review?")) {
umf = efopen("unmerged", "w");
elseek(mfd, 0L, 0);
while(eread(mfd, &newprs, 4)) {
if(!newprs) continue;
p = select_person(newprs);
if(!p) continue;
s = datestring(p->birth);
fullbirthname(newprs, name1);
fprintf(umf, "%ld: %s %s\n", newprs, name1, s);
}
fclose(umf);
printf("See the file `unmerged' in your gng directory.\n");
}
}
} /* mergestats */

merge_current()
{
long prs2;
long save_prs = cur_prs_rowid;

/* no merge file descriptor, because this merge does not come from a foldin command */
mfd = -1;
printf("\nPerson to merge with %s %s?", cur_fname, cur_lname);
prs2 = searchfor(NULL);
if(prs2 == cur_prs_rowid) {
printf("This person is the same as the current person, already `merged'.\n");
printf("Shall I continue the scan/merge process with %s's relatives?", cur_fname);
if(!get_yn("")) return;
nl();
diffbufset();
merge_relatives(cur_prs_rowid);
} else {
nl();
diffbufset();
mergewith(cur_prs_rowid, prs2);
}

diffbufclear();
set_curperson(save_prs);
} /* merge_current */

/* merge two people; indirectly recursive */
mergewith(prs1, prs2)
long prs1, prs2;
{
int parentflags;
char change1, change2, choice;
struct person p1, p2;

ia_flag = 1;
fullbirthname(prs1, name1);
fullbirthname(prs2, name2);
printf("Merging %s", name1);
if(strcmp(name1, name2)) printf(" with %s", name2);
nl();
nl();

parentflags = merge_parenttest(prs1, prs2);
if(parentflags&0x44) {
printf("Merging these two individuals would create one person with two %sthers.\n\n",
parentflags&4 ? "fa" : "mo");
longjmp(main_menu, 1);
}

/* person records, resolve sex and dates */
select_copy_person(prs1, &p1);
select_copy_person(prs2, &p2);
if(p1.sex != p2.sex) {
printf("These two individuals are not the same sex, cannot merge.\n");
return;
}

choice = mergedate("birthdate", p1.birth, p2.birth);
if(choice == 1) {
p2.birth = p1.birth;
change2 = 1;
}
if(choice == 2) {
p1.birth = p2.birth;
change1 = 1;
}

choice = mergedate("deathdate", p1.death, p2.death);
if(choice == 1) {
p2.death = p1.death;
change2 = 1;
}
if(choice == 2) {
p1.death = p2.death;
change1 = 1;
}

if(change1) insupd_person(prs1, &p1);
if(change2) insupd_person(prs2, &p2);

mergename(prs1, prs2);
mergenote(prs1, prs2);
mergetext(prs1, prs2);
if(p1.sex == 'f')
mergelink(prs1, prs2, 'h', "husband");
else
mergelink(prs1, prs2, 'w', "wife");
mergelink(prs1, prs2, 'c', "child");
mergelink(prs1, prs2, 'm', "mother");
mergelink(prs1, prs2, 'f', "father");

/* now drop person 2 */
delperson(prs2);
++mergecount;

merge_relatives(prs1);
} /* mergewith */

/* see if the merge would create multiple parents */
merge_parenttest(prs1, prs2)
long prs1, prs2;
{
int rc; /* return code */
long parent1, parent2;

rc = 0;

if(parent1 = getdad(prs1))
rc |= 1;
if(parent2 = getdad(prs2))
rc |= 2;
if((rc&3) == 3 && parent1 != parent2) rc |= 4;

if(parent1 = getmom(prs1))
rc |= 16;
if(parent2 = getmom(prs2))
rc |= 32;
if((rc&48) == 48 && parent1 != parent2) rc |= 64;

return rc;
} /* merge_parenttest */

mergedate(datename, d1, d2)
char *datename;
date d1, d2;
{
int exact1, exact2;
long diff;
int sameyear;

diff = d1-d2;
if(diff < 0) diff = - diff;
sameyear = 0;
if(d1/YEAR_DAY_RATIO == d2/YEAR_DAY_RATIO) sameyear = 1;
exact1 = d1%YEAR_DAY_RATIO;
exact2 = d2%YEAR_DAY_RATIO;

if(d1 == d2) {
if(!ia_flag) {
if(exact1) return 6; /* exact match on dates */
if(d2) return 4; /* match on inexact dates */
}
return 0; /* match on null */
} /* equal dates */

/* perhaps one is null */
if(!d1) return ia_flag ? 2 : 0;
if(!d2) return ia_flag ? 1 : 0;

/* perhaps they aren't even close */
if(!ia_flag || diff > 9*YEAR_DAY_RATIO) {
char choice;
diffdates:
if(!ia_flag) {
if(diff > 10*YEAR_DAY_RATIO) return -4;
if(diff > 4*YEAR_DAY_RATIO) return 0;
if(sameyear && exact1|exact2) {
if(exact1 && exact2)
return 2;
return 4;
}
if(!(diff%YEAR_DAY_RATIO) && exact1) return 4;
if(exact1 && exact2) return -4;
return 0;
}

printf("1 : %s\n", datestring(d1));
printf("2 : %s\n", datestring(d2));
printf("Select a ");
choice = promptfor(datename, "12");
nl();
nl();
return choice - '0';
} /* different */

/* is one exact while the other is not? */
if(!exact2 && exact1) return 1;
if(exact2 && !exact1) return 2;
goto diffdates;
} /* mergedate */

namecmp(full1, full2)
char *full1, *full2;
{
char fname[NAMELEN], mname[NAMELEN], lname[NAMELEN];
char extendflags;

if(!strcmp(full1, full2)) return 3;

splitstring(full1);
strcpy(fname, fld1);
strcpy(mname, fld2);
strcpy(lname, fld3);
splitstring(full2);
extendflags = 0;

extendflags |= namecmp1(fname, fld1);
extendflags |= namecmp1(mname, fld2);
extendflags |= namecmp1(lname, fld3);
if(extendflags & 4) return 0; /* no match */
if((extendflags&3) == 3) return 0;
return extendflags;
} /* namecmp */

namecmp1(s1, s2)
char *s1, *s2;
{
int l1, l2, d, min, rc;

l1 = strlen(s1);
l2 = strlen(s2);

if(l1 == l2) {
if(strcmp(s1, s2)) return 4;
return 0;
}

if(!l1) return 2;
if(!l2) return 1;

if(l1 < l2) {
rc = 2;
min = l1;
d = l2-l1;
} else {
rc = 1;
min = l2;
d = l1-l2;
}
if(!strncmp(s1, s2, min)) return rc;

if(min >= 4) {
min = (min*3+1) / 4;
if(!strncmp(s1, s2, min)) return rc;
}
return 4;
} /* namecmp1 */

mergename(prs1, prs2)
long prs1, prs2;
{
struct name n1, n2, *n;
long n1rowid;
int namecnt, i, cmpf;
int match_cnt, extend_cnt;
long bn1, bn2, bn1_to, bn2_to;
char choice;
struct rel_cursor c;

match_cnt = extend_cnt = 0;
bn1 = bn2 = bn1_to = bn2_to = 0;

namecnt = 1;
while(1) {
i = 0;
IDX_LOOP_name(&c, n, &n1, prs1, "prsid")
if(++i == namecnt) break;
END_LOOP_name(&c)
if(!n) break;

n1rowid = c.rowid;
if(n1.birth) bn1 = c.rowid;
++namecnt;

IDX_LOOP_name(&c, n, &n2, prs2, "prsid")
if(n2.birth) bn2 = c.rowid;
cmpf = namecmp(n1.fullname, n2.fullname);
if(!cmpf) continue; /* different names */

if(n1.birth) bn1_to = c.rowid;
if(n2.birth) bn2_to = n1rowid;

if(cmpf == 3) { ++match_cnt; break; }

/* one is an extension of the other */
++extend_cnt;
if(!ia_flag) break;
if(cmpf == 1) break;

strcpy(n1.fullname, n2.fullname);
gng_hashname(&n1);
insupd_name(n1rowid, &n1);
break;
END_LOOP_name(&c)

if(ia_flag && n) { /* name match */
if(c.rowid != bn2 &&
c.rowid != bn1_to)
delete_name(c.rowid);
}

} /* loop over names for prs1 */

if(!ia_flag) return match_cnt*7 + extend_cnt*5;

/* resolve different birth names */
if(!bn1) eprint("@mergename finds no birthname for person %ld", prs1);
if(!bn2) {
birthname(prs2);
bn2 = birth_rowid;
}

if(bn1_to != bn2) {
/* birthnames do not match up */
select_copy_name(bn1, &n1);
select_copy_name(bn2, &n2);

/* see if one links up to a secondary name, while the other does not */
if(bn1_to && !bn2_to) {
n1.birth = 0;
insupd_name(bn1, &n1);
n2.prsid = prs1;
insupd_name(NULLKEY, &n2);
} else if(bn2_to && !bn1_to) {
n2.birth = 0;
insupd_name(bn2, &n2);
n1.prsid = prs2;
insupd_name(NULLKEY, &n1);
} else {
splitstring(n1.fullname);
printf("1 : %s %s %s\n", fld1, fld2, fld3);
splitstring(n2.fullname);
printf("2 : %s %s %s\n", fld1, fld2, fld3);
choice = promptfor("Select a birth (maiden) name", "12");
nl();
nl();
choice -= '0';
/* the unselected loses its birth status */
if(choice == 1) {
n2.birth = 0;
insupd_name(bn2, &n2);
if(bn1_to) {
select_copy_name(bn1_to, &n2);
n2.birth = 1;
insupd_name(bn1_to, &n2);
bn1_to = 0;
} else {
n1.prsid = prs2;
insupd_name(NULLKEY, &n1);
}
} else {
n1.birth = 0;
insupd_name(bn1, &n1);
if(bn2_to) {
select_copy_name(bn2_to, &n1);
n1.birth = 1;
insupd_name(bn2_to, &n1);
bn2_to = 0;
} else {
n2.prsid = prs1;
insupd_name(NULLKEY, &n2);
}
}
}
if(bn1_to) delete_name(bn1_to);
if(bn2_to) delete_name(bn2_to);
}

/* transfer names to prs1 */
IDX_LOOP_name(&c, n, &n2, prs2, "prsid")
if(n2.birth) continue;
n2.prsid = prs1;
insupd_name(NULLKEY, &n2);
delete_name(c.rowid);
END_LOOP_name(&c)
} /* mergename */

/* only called when ia_flag is set */
mergenote(prs1, prs2)
long prs1, prs2;
{
int i, l1, l2;
char choice;
struct rel_cursor c;
struct qknote *qn;
struct qknote list1[QKNOTE_CAT], list2[QKNOTE_CAT];
long rows1[QKNOTE_CAT], rows2[QKNOTE_CAT];

for(i=0; i<QKNOTE_CAT; ++i)
rows1[i] = rows2[i] = 0;

IDX_LOOP_qknote(&c, qn, NULL, prs1, "prsid")
i = strchr(qkcat, qn->ntype) - qkcat;
rows1[i] = c.rowid;
list1[i] = *qn; /* structure copy */
END_LOOP_qknote(&c)

IDX_LOOP_qknote(&c, qn, NULL, prs2, "prsid")
i = strchr(qkcat, qn->ntype) - qkcat;
rows2[i] = c.rowid;
list2[i] = *qn; /* structure copy */
END_LOOP_qknote(&c)

for(i=0; i<QKNOTE_CAT; ++i) {
if(!rows2[i]) continue; /* no second note */
if(!rows1[i]) {
copynote:
list2[i].prsid = prs1;
insupd_qknote(rows1[i], &list2[i]);
continue;
}

/* both notes present, take one if it is significantly longer */
l1 = strlen(list1[i].note);
l2 = strlen(list2[i].note);
if(l1 - l2 > 20) continue;
if(l2 - l1 > 20) goto copynote;
if(!strcmp(list1[i].note, list2[i].note)) continue;

/* different notes of similar length, ask the user */
printf("Different entries for %s\n", qkdesc(qkcat[i], 1));
printf("1 : %s\n", list1[i].note);
printf("2 : %s\n", list2[i].note);
choice = promptfor("Select a short note", "12");
nl();
nl();
if(choice == '2') goto copynote;
} /* end loop over notes */
} /* mergenote */

/* before we adjust points based on common links, a moment's reflection.
 * A point system, table driven, easy to code, and very flexible,
 * suffers from the implicit assumption of statistical independence.
 * Each event contributes x points, regardless of the surrounding events.
 * In practice this is not always true, as illustrated by twins.
 * Same last name and birthdate, same parents,
 * and if they went through the foster care system,
 * many of the same caregiving parents.
 * Yet each new common foster parent does not give much additional
 * credibility to the assertion that these two individuals
 * are in fact the same person.
 * Furthermore, the same last name doesn't add much to the confidence level,
 * once we have established the same parent.
 * We reduce common parents to a yes or no item;
 * doesn't matter how many caregiving parents you have in common.
 * Similarly we reduce common children to yes or no.
 * When comparing names of imediate relatives, we only look at first names,
 * since last names are likely to match your  last name,
 * which has already been compared.
 *
 * There are unfortunately two contexts within which
 * the evaluation routine is required to run.
 * In the first scenario, two people are drawn from the population of your
 * database, or perhaps the entire world, and compared.
 * In the second, two people are known to be related to a third
 * person, usually the result of a prior merge,
 * and we are asking whether these two individuals are the same.
 * In the former case a common first name is almost worthless,
 * but in the latter case a common first name is almost a guaranteed match.
After all, how many parents give two children the same name?
 * And with the disclaimer on twins, common birthdate is worth
 * a great deal, given a common parent.
 * When a common link exists, each piece of data
 * is magnified, contributing considerably more points
 * towards our confidence level,
 * because the population is so much smaller.
 * Assigning a relatively large number of points to a common link helps,
 * but does not really solve the problem.
 * For instance, comon birth year is worth almost
 * as much as common birthdate once a common parent is established.
 * Two separate deliveries in one year is quite rare.
 * But within the pool of a large database,
 * common birthdate is 365 times as valuable as comon birth year.
 * So we either differentiate between these two contexts,
 * or slant our algorithms towards one of them.
 * Being lazy, I take the easy way out and slant towards the latter context.
 * Except for the first merge, which could be designated by the human if
 * necessary, almost all merges result from folding in a family tree;
 * merge the relatives of a person who has just been merged.
 * Therefore this system "assumes" there is a common relative.
 * To safely deal with the "large pool" merge,
 * we require lots of similar relatives by name and birthdate,
 * more than you might first expect, since I wouldn't want to match two
 * people because they each have three children with the same birth years
 * and similar names.
 * Besides, you only have to find one such match out of hundreds
 * to begin folding in a family tree,
 * and after that all merges have common relatives.
 * There is usually some oblidging soul with lots of spouses
 * and children in any family tree.
 * As mentioned earlier, I give lots of points to a common link,
 * in a meager attempt to make the two contexts compatible.
 * A common link is worth the equivalent of 4 name/birthdate matches.
 * Thus having common children is equivalent to having four children,
 * same names and birthdates. */

static struct LDATA {
long rownum, prsid;
date ls, le;
date bd;
char fname[7];
char adopt;
};

mergelink(prs1, prs2, linktype, relation)
long prs1, prs2;
char linktype;
char *relation;
{
int n1, n2, i1, i2;
int newpoints, totpoints = 0;
char choice;
struct prslink new_l;
struct LDATA ld1[SPREADMAX];
struct LDATA ld2[SPREADMAX];
char marstring[NAMELEN+50];

n1 = ldata_setup(prs1, prs2, linktype, ld1, 1);
n2 = ldata_setup(prs2, prs1, linktype, ld2, 0);

for(i2=0; i2<n2; ++i2) {
newpoints = 0;
for(i1=0; i1<n1; ++i1) {

if(ld1[i1].prsid != ld2[i2].prsid) {
int bdval, fnval;

if(ia_flag) continue;
/* maybe these are the same people, though not yet merged */
/* first check the birthdates */
bdval = mergedate("x", ld1[i1].bd, ld2[i2].bd);
if(bdval < 0) continue; /* forget it */
fnval = 0;
if(!strncmp(ld1[i1].fname, ld2[i2].fname, 4)) fnval = 5;
if(!strncmp(ld1[i1].fname, ld2[i2].fname, 7)) fnval = 7;
bdval += fnval;
if(linktype == 'h' || linktype == 'w')
bdval += mergedate("x", ld1[i1].ls, ld2[i2].ls);
if(bdval > newpoints) newpoints = bdval;
continue;
}

/* same link, must merge */
birthname(ld1[i1].prsid);
if(ld1[i1].adopt != ld2[i2].adopt) {
printf("Cannot make %s %s the %s of both individuals,\n\
as one link is a blood relation and the other is a caregiving relation.\n",
fld1, fld3, relation);
longjmp(main_menu, 1);
}

if(!ia_flag) {
newpoints = 52;
if(linktype == 'h' || linktype == 'w')
newpoints += mergedate("x", ld1[i1].ls, ld2[i2].ls);
return newpoints;
}

select_copy_prslink(ld1[i1].rownum, &new_l);
new_l.from1 = new_l.from2 = 0;

if(linktype == 'h' || linktype == 'w') {
sprintf(marstring, "wedding date (%s %s)", fld1, fld3);
choice = mergedate(marstring, ld1[i1].ls, ld2[i2].ls);
if(choice == 2) {
up_aniv(ld1[i1].prsid, prs1, ld1[i1].ls, ld2[i2].ls);
up_aniv(prs1, ld1[i1].prsid, ld1[i1].ls, ld2[i2].ls);
new_l.start = ld2[i2].ls;
}

sprintf(marstring, "divorce date (%s %s)", fld1, fld3);
choice = mergedate(marstring, ld1[i1].le, ld2[i2].le);
if(choice == 2) {
new_l.end = ld2[i2].le;
}
}

if(ld1[i1].adopt) {
sprintf(marstring, "date for start of caregiving relationship (%s %s)", fld1, fld3);
choice = mergedate(marstring, ld1[i1].le, ld2[i2].le);
if(choice == 2) {
new_l.end = ld2[i2].le;
}
}

insupd_prslink(ld1[i1].rownum, &new_l);

/* don't need the old link any more */
delete_prslink(ld2[i2].rownum);
break;
} /* end loop comparing all linked relatives of prs1 to a specific linked relative of prs2 */

if(!ia_flag) {
totpoints += newpoints;
continue;
}

if(i1 < n1) continue;

/* went through and didn't find this person, transfer the link */
select_copy_prslink(ld2[i2].rownum, &new_l);
new_l.from1 = 0;
new_l.from2 = 1;
if(linktype == 'w' || linktype == 'c')
new_l.prs1 = prs1;
else new_l.prs2 = prs1;
insupd_prslink(ld2[i2].rownum, &new_l);
}

return totpoints;
} /* mergelink */

ldata_setup(prsid, otherid, linktype, array, set_1)
long prsid, otherid;
char linktype;
struct LDATA *array;
{
struct rel_cursor c;
struct prslink *l, new_l;
int n;
long linked_to;
char *idxname;

idxname = "prs2";
if(linktype == 'w' || linktype == 'c')
idxname = "prs1";
if(linktype == 'w') linktype = 'h';

n = 0;
IDX_LOOP_prslink(&c, l, &new_l, prsid, idxname)
if(new_l.linktype != linktype) {
if(linktype != 'c') continue;
if(new_l.linktype == 'h') continue;
}
array[n].prsid = linked_to = new_l.prs1 + new_l.prs2 - prsid;
if(linked_to == otherid)
eprint("Merging immediate relatives %ld and %ld, person cannot become his own parent", prsid, otherid);
array[n].rownum = c.rowid;
array[n].adopt = new_l.adopt;
array[n].ls = new_l.start;
array[n].le = new_l.end;

if(!ia_flag) {
birthname(linked_to);
/* I assume this will not write past the 7th byte */
strncpy(array[n].fname, fld1, 7);
array[n].bd = (select_person(linked_to))->birth;
++n;
continue;
}

++n;
if(!set_1) continue;
if(new_l.from1 && !new_l.from2) continue;
new_l.from1 = 1;
new_l.from2 = 0;
insupd_prslink(c.rowid, &new_l);
END_LOOP_prslink(&c)
return n;
} /* ldata_setup */

/* evaluate merge probability, are these people likely to be the same? */
/* returns a confidence level from 0 to 3 */
evaluate(prs1, prs2, common_relat, printval)
long prs1, prs2;
int common_relat; /* known to have common parent or descendent */
{
struct person p1, p2;
int totpoints = 0;
int rc;

ia_flag = 0;

if(printval) /* debug */
printf("eval %ld/%ld", prs1, prs2);

/* start by checking sex, birthdate, deathdate, and names. */
select_copy_person(prs1, &p1);
select_copy_person(prs2, &p2);
/* don't even compare people of different sexes */
if(p1.sex == p2.sex) {

totpoints += mergedate("x", p1.birth, p2.birth);
totpoints += mergedate("x", p1.death, p2.death);
totpoints += mergename(prs1, prs2);

if(common_relat) totpoints += 52;
if(printval) printf(",%d", totpoints);

if(p1.sex == 'f')
totpoints += mergelink(prs1, prs2, 'h', "husband");
else
totpoints += mergelink(prs1, prs2, 'w', "wife");
if(printval) printf(",%d", totpoints);
totpoints += mergelink(prs1, prs2, 'c', "child");
} /* sex check */

if(printval) printf(",%d\n", totpoints);

/* if we have two links, and common first name,
 * that's a certain match */
if(totpoints >= 52*2+5) return 5;

/* one link and lots of children by the same name is great */
if(totpoints >= 52+7*5) return 4;

/* one linke and a common first name and birthyear is good enough
 * for an automatic merge, if nobody else lies
 * in the same range. */
if(totpoints >= 52+7+4) return 3;

/* One linke and a name or birthdate is almost as good.
 * This assumes one database will not know about twin-a
 * while another database knows about twin-b.
 * Usually the knowledge of one database is a subset
 * of the knowledge of another.
 * Hence twin-a will match both twin-a and twin-b,
 * and the program will ask you who you want to merge with. */
if(totpoints >= 52+6) return 2;

/* common name/birthdate means we should consider the merge. */
if(totpoints >= 30) return 1;

/* nothing in common */
return 0;
} /* evaluate */

merge_relatives(base_prs)
long base_prs;
{
if((select_person(base_prs))->sex == 'f')
mergerel1(base_prs, 'h');
else
mergerel1(base_prs, 'w');
mergerel1(base_prs, 'c');
/* we need these two lines, for caregiving parents */
mergerel1(base_prs, 'f');
mergerel1(base_prs, 'm');
} /* merge_relatives */

mergerel1(base_prs, linktype)
long base_prs;
char linktype;
{
struct rel_cursor c;
struct prslink *l;
int n;
long bestprs[3];
int bestcnt, bestval;
long cur_prs, second_prs, cur_lrow;
char val, dospread;
char cur_adopt, temp_adopt, cur_from;
char cmd, orig_linktype;
char *idxname, *digits;

idxname = "prs2";
if(linktype == 'w' || linktype == 'c')
idxname = "prs1";
orig_linktype = linktype;
if(linktype == 'w') linktype = 'h';

while(1) {
IDX_LOOP_prslink(&c, l, NULL, base_prs, idxname)
if(l->linktype != linktype) {
if(linktype != 'c') continue;
if(l->linktype == 'h') continue;
}
if(l->from1 | l->from2) break;
END_LOOP_prslink(&c)
if(!l) break;

cur_from = 1;
if(l->from2) cur_from = 2;
cur_adopt = l->adopt;
cur_prs = l->prs1 + l->prs2 - base_prs;
cur_lrow = c.rowid;

/* now loop through links looking for somebody
 * who resembles cur_prs */
bestval = bestcnt = 0;
n = -1;
dospread = 1;
while(1) {
if(dospread)
spreadout(base_prs, orig_linktype);
dospread = 1;
if(++n >= spreadnum) break;

dospread = 0;
l = select_prslink(spreaddata[n].rownum);
if(cur_from == 1 && !l->from2) continue;
if(cur_from == 2 && !l->from1) continue;
temp_adopt = l->adopt;
if(temp_adopt != cur_adopt) continue;
dospread = 1;

second_prs = spreaddata[n].prsid;
if(indiffbuf(cur_prs, second_prs)) continue;
val = evaluate(cur_prs, second_prs, (linktype == 'c'), 0);
if(!val) continue;
if(val < bestval) continue;

if(val > bestval) {
/* up the stakes */
bestcnt = 0;
bestval = val;
}
if(bestcnt < 3)
bestprs[bestcnt++] = second_prs;
} /* end loop looking for a match */

if(!bestval) goto linkoff;
if(bestcnt == 1 && bestval > 1) {
second_prs = bestprs[0];
goto mergego;
}

birthname(cur_prs);
if(bestcnt > 1)
printf("%s %s resembles %d different people, I need your help.\n",
fld1, fld3, bestcnt);
else
printf("Tenuous match, I don't know whether to run this merge.\n");
printf("Here is the summary information on %s %s:", fld1, fld3);
set_curperson(cur_prs);
topscreen(stdout, 0);
printf("                    CANDIDATE %s:",
bestcnt > 1 ? "1" : "MATCH");
set_curperson(bestprs[0]);
topscreen(stdout, 0);
if(bestcnt > 1) {
printf("                    CANDIDATE 2:");
set_curperson(bestprs[1]);
topscreen(stdout, 0);
digits = "012";
}
if(bestcnt > 2) {
printf("                    CANDIDATE 3:");
set_curperson(bestprs[2]);
topscreen(stdout, 0);
digits = "0123";
}

if(bestcnt > 1) {
++mergepromptcount;
cmd = promptfor("Which candidate should I merge with the original person (0 for none)?", digits);
nl();
nl();
if(cmd == '0') {
int i;
struct prslink new_l;

linkoff:
for(i=0; i<bestcnt; ++i)
diffbufput(cur_prs, bestprs[i]);

select_copy_prslink(cur_lrow, &new_l);
new_l.from1 = new_l.from2 = 0;
insupd_prslink(cur_lrow, &new_l);
continue;
}

++mergeyescount;
cmd -= '1';
second_prs = bestprs[cmd];

mergego:
/* either merge these two, or some common ancestor of these two */
if(merge_anc(cur_prs, second_prs)) continue;

fullbirthname(cur_prs, name1);
fullbirthname(second_prs, name2);
printf("Could not merge %s with %s.\n", name1, name2);
printf("Although these two individuals appear to be the same,\n\
They have different ancestors, parents with very little in common.\n\
Please investigate this situation.\n\n");
bestcnt = 0;
goto linkoff;
}

++mergepromptcount;
cmd = get_yn("Shall I merge these two people?");
nl();
if(!cmd) goto linkoff;
++mergeyescount;
second_prs = bestprs[0];
goto mergego;
} /* end loop over mergable relatives */
} /* mergerel1 */

merge_anc(prs1, prs2)
long prs1, prs2;
{
int parentflags;
long anc1, anc2;

parentflags = merge_parenttest(prs1, prs2);
if(!(parentflags&0x44)) {
mergewith(prs1, prs2);
return 1;
}

if(parentflags & 4) { /* multiple fathers */
anc1 = getdad(prs1);
anc2 = getdad(prs2);
if(evaluate(anc1, anc2, 1, 0) >= 2)
return merge_anc(anc1, anc2);
} /* two fathers */

if(parentflags & 0x40) { /* multiple mothers */
anc1 = getmom(prs1);
anc2 = getmom(prs2);
if(evaluate(anc1, anc2, 1, 0) >= 2)
return merge_anc(anc1, anc2);
} /* two mothers */

return 0;
} /* merge_anc */

merge_search()
{
struct rel_cursor c;
long newprs, origprs;
struct person *p;
struct name *n;
long hashval;
int i;
long list[10];

nl();
origprs = 0;

elseek(mfd, 0L, 0);
top:
while(eread(mfd, &newprs, 4)) {
if(!newprs) continue;
p = select_person(newprs);
if(!p) continue;

/* find other people having the same hashed name */
birthname(newprs);
hashval = (select_name(birth_rowid))->hashval;
i = 0;
IDX_LOOP_name(&c, n, NULL, hashval, "hashval")
if(i == 10) break;
if(n->prsid == newprs) continue;
list[i++] = n->prsid;
END_LOOP_name(&c)

/* evaluate each candidate */
for(--i; i>=0; --i) {
if(evaluate(list[i], newprs, 0, 0) < 5) continue;
origprs = list[i];
break;
}

if(origprs) break;
} /* scan through mfd file */

diffbufset();
if(!origprs) {
printf("Sorry, I cannot crossmatch any of the new people with any of your original people.\n");
} else {
if(!merge_anc(origprs, newprs)) goto top;
}
diffbufclear();
} /* merge_search */

static mergetext(prs1, prs2)
long prs1, prs2;
{
char *fn;
static char remember_side, remask;
char side, aside, txtfound;
int year;
char *s;
char prs1str[20];
char prs2str[20];
int prs1len, prs2len;

sprintf(prs1str, "text/%ld.", prs1);
prs1len = strlen(prs1str);
sprintf(prs2str, "text/%ld.", prs2);
prs2len = strlen(prs2str);
side = 0;
if(remember_side) side = remember_side;
txtfound = 0;

dirstart = 1;
while(fn = gng_nextfile()) {
if(strncmp(fn, prs2str, prs2len)) continue;
txtfound = 1;
s = fn+prs2len;
/* do we already have text entries for this person? */
strcpy(prs1str+prs1len, s);
if(!strcmp(s, "Y__")) year = -1;
else year = atoi(s+1);
if(access(prs1str, 0))
aside = 2; /* no conflicting text entries */
else {
/* entries for both people, which one do you want? */
if(side) aside = side;
else {
printf("Sorry, intelligent text entry merge not yet implemented.\n\
You must decide whether to retain the original text entries,\n\
or overwrite them with the incoming text entries.\n");
if(get_yn("Shall I give precedence to the new, incoming text entries?"))
side = 2;
else
side = 1;
aside = side;
if(!remask) {
if(get_yn("Shall I apply this rule to all subsequent merges?"))
remember_side = side;
remask = 1;
}
nl();
}
}

if(aside == 1) {
unlink(fn);
casc_txt(prs2, NULLKEY, year);
} else {
unlink(prs1str);
casc_txt(prs1, NULLKEY, year);
erename(fn, prs1str);
casc_txt(prs2, prs1, year);
}
} /* loop over textfiles */

if(txtfound) {
struct person p;
select_copy_person(prs1, &p);
p.txtc1 = 0; /* recalculate text entries */
insupd_person(prs1, &p);
}
} /*mergetext */

/* like casc_upd_del, but there is no index on textent.prsid */
static casc_txt(prs1, prs2, year)
long prs1, prs2;
{
struct textent te, *t;
struct rel_cursor c;

LOOP_textent(&c, t, &te)
if(te.prsid != prs1) continue;
if(year < 0) {
if(te.when) continue;
} else {
if(!te.when) continue;
if((te.when/YEAR_DAY_RATIO)%100 != year) continue;
}

if(prs2) {
te.prsid = prs2;
insupd_textent(c.rowid, &te);
} else {
delete_textent(c.rowid);
}
END_LOOP_textent(&c)
} /* casc_txt */

