/***********************************************************************
|  Permutation group analyzer
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* grpnode.ec: structures to hold and connect (chain together) permutations */

/* compile large model, and link /stack:8000 */

/*-----------------------------------------------------------------------------
This program produces a set of strong generators for a permutation group
and prints the order of the group.
It then accepts permutations from stdin
and writes each using the strong/given generators.
See /math/info/group[perm_grp] for a description of the algorithms
used in this program.

If x appears in position i in a permutation,
then element i (or the element in slot i) is moved to
position x.Therefore the successive application of permutations a and b is
c[i] = b[a[i]].
As a regression test, run grpanal rubcube3,
which analyzes the permutation group of the standard Rubik cube.

the order of the group is 24*24*21*22*20*18*18*15*16*14*12*12*9*10*8*6*6*2
-----------------------------------------------------------------------------*/


${
/* copy a permutation, syntax like strcpy or memcpy */
#define permcpy(p1, p2) memcpy(p1, p2, PERMLEN);
#define NUM_MOVES 50
#define PATHLEN 1000

extern short permsize; /* size of permutations */
extern short pathlen;
extern short remap_mode;
/* nontrivial coset reps at level n, link lists */
extern struct list_perm cosrep[];
extern struct perm *idn; /* identity permutation */
/* the initial moves */
extern short num_moves, num_rots;
extern char movename[];
extern short *moveperm[];
extern short *rots[];

/* functions with non-integer returns */
extern struct perm *find_cosrep();
$}

global:
usage "usage: grpanal [-d] [-s#] perm_grp_desc_file <configurations";

main:
arguments ".":
options "dqs:";

node list order_by "." connect {
struct perm *strong; /* strong generator */
short given; /* a given generator */
short invert; /* 1 if you should invert this perm */
}:
/* a trail of strong generators used to reach the permutation in tp3.
set up this list before calling fold_in_gen().
This list is then saved before calling cross(),
which calls fold_in_gen() recursively.
Add_cr() copies this list to record how the coset rep was produced. */
gvar trail;

/* store a permutation as an array of shorts */
node list order_by "." perm {
#define PERMSIZE 96 /* max number of elements being permuted */
#define PERMLEN (sizeof(short) * permsize)
short p[PERMSIZE]; /* the permutation */
/* build this permutation from strong/given generators */
struct list_connect produce;
/* attribute bits */
#define CR_INGROUP 1 /* part of the group under consideration */
#define CR_NOTSEED 2 /* coset rep is more than just a seed */
#define CR_SUBGROUP 4 /* part of the subgroup being built */
#define CR_PENDING 8 /* pending generator */
short attrib; /* attributes */
/* it is not tremendously difficult to overflow this long, I don't check for this */
long tostart; /* length of path back to start */
};

${
short permsize; /* size of permutations */
short pathlen;
short remap_mode;
/* nontrivial coset reps at level n, link lists */
struct list_perm cosrep[PERMSIZE];
struct perm *idn; /* identity permutation */
/* the initial moves */
short num_moves, num_rots;
char movename[NUM_MOVES];
short *moveperm[NUM_MOVES];
short *rots[NUM_MOVES];
$}

